//  IEEE1888 / IEEE1888 Downloader Library for Arduino
//
//  2012/03/07 ver.1   H.Ochiai
//

// --------- IEEE1888FetchClient.cpp (begin) ---------
#include <SPI.h>
#include <Ethernet.h>
#include <EthernetClient.h>
#include <avr/pgmspace.h>     // Retrieve Strings from the Program Memory
#include "IEEE1888FetchClient.h"

static char NULL_CHARACTER = '\0';

// Initialize the IEEE1888FetchClient instance
void IEEE1888FetchClient::begin(
		  const uint8_t* server_ip4,
		  const char* server_host,
		  const char* server_path,
		  uint16_t server_port,
		  fetch_deal _fetch_deal){

	this->server_ip4=server_ip4;
	this->server_host=server_host;
	this->server_path=server_path;
	this->server_port=server_port;
	this->_fetch_deal = _fetch_deal;
}


// Messages (Stored in the program memory) -- HTTP Header Part
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER01[] =  "POST ";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER02[] =  " HTTP/1.1";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER03[] =  "Content-Type: text/xml charset=UTF-8";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER04[] =  "User-Agent: IEEE1888FetchClient (Arduino)";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER05[] =  "Host: ";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER06[] =  "SOAPAction: \"http://soap.ieee1888.org/query\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPHEADER07[] =  "Content-Length: ";

// Messages (Stored in the program memory) -- HTTP Body Part
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY01[] = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY02[] = "<soapenv:Envelope xmlns:soapenv=";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY03[] = "\"http://schemas.xmlsoap.org/soap/envelope/\">";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY04[] = "<soapenv:Body>";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY05[] = "<ns2:queryRQ xmlns:ns2=\"http://soap.ieee1888.org/\">";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY06[] = "<transport xmlns=\"http://www.biigroup.com\">";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY07[] = "<header>";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY08[] = "<query id=\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY09[] = "\" type=\"storage\">";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY10[] = "<key id=\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY11[] = "\" attrName=\"time\" ";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY12[] = "select=\"maximum\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY13[] = "select=\"minimum\"";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY14[] = "lt=\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY15[] = "\" ";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY16[] = "gt=\"";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY17[] = "\"";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY18[] = "/>";

PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY19[] = "</query></header></transport></ns2:queryRQ>";
PROGMEM prog_char IEEE1888FetchClient_Get_HTTPBODY20[] = "</soapenv:Body></soapenv:Envelope>";


int IEEE1888FetchClient::get(fetch_point p[], int pCount){
  int ret_code = IEEE1888_DOWNLOAD_OK;
  int rescode = 0;  // HTTP response code
  int clen = 0;     // content length
  char count;
  unsigned char c;

  //char * query_id = "ac34cc67-1234-4567-3456-dd45bcac8900";
  char * query_id = (char*)malloc(36);

  EthernetClient client;
  if (!client.connect(server_host, server_port)) 
  {
	Serial.print("connect fail!!!!!");
    return(IEEE1888_DOWNLOAD_CONNFAIL);
  }

  //query_id = ieee1888_mk_new_uuid(); // 获取唯一标识query的uuid
  ieee1888_mk_new_uuid(query_id);
  Serial.println(query_id);
  
  
  clen = 375; /*这个值固定，代表strlen(IEEE1888FetchClient_Get_HTTPBODY01)+....strlen(IEEE1888FetchClient_Get_HTTPBODY09)+strlen(IEEE1888FetchClient_Get_HTTPBODY19)
				+strlen(IEEE1888FetchClient_Get_HTTPBODY20)+strlen(query_id)+2个回车换行的长度*/
  int i = 0;
  for(i; i<pCount; i++)
  {
	if(p[i].select == 0 || p[i].select == 1) // 获取最新值
		clen += 16;	  //16表示strlen(IEEE1888FetchClient_Get_HTTPBODY12)或strlen(IEEE1888FetchClient_Get_HTTPBODY13)
	else if(p[i].select == 2)
		clen += 11 + strlen(p[i].pGt) + strlen(p[i].pLt); //11表示strlen(IEEE1888FetchClient_Get_HTTPBODY14)+...+strlen(IEEE1888FetchClient_Get_HTTPBODY17)
	clen += 29; //strlen(IEEE1888FetchClient_Get_HTTPBODY10)+strlen(IEEE1888FetchClient_Get_HTTPBODY11)+strlen(IEEE1888FetchClient_Get_HTTPBODY18)
	clen += strlen(p[i].pid);
  }

  // sending message buffer
  char sbuf[80];

  // send HTTP header
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER01);
  client.print(sbuf);  // "POST "
  client.print(server_path);
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER02);
  client.println(sbuf);  // " HTTP/1.1"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER03);
  client.println(sbuf);  // "Content-Type: text/xml; charset=UTF-8"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER04);
  client.println(sbuf);  // "User-Agent: IEEE1888FetchClient (Arduino)"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER05);
  client.print(sbuf);  // "Host: "
  client.println(server_host);
  // client.println("Content-Type: application/soap+xml; charset=UTF-8");
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER06);
  client.println(sbuf);  // "SOAPAction: \"http://soap.1888.org/query\""
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPHEADER07);
  client.print(sbuf);  // "Content-Length: "
  client.println(clen);
  client.println();

  // send HTTP body
  client.println();
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY01);
  client.print(sbuf);  // "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY02);
  client.print(sbuf);  // "<soapenv:Envelope xmlns:soapenv="
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY03);
  client.print(sbuf);  // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY04);
  client.print(sbuf);  // "<soapenv:Body>"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY05);
  client.print(sbuf);  // "<ns2:queryRQ xmlns:ns2=\"http://soap.ieee1888.org/\">" 
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY06);
  client.print(sbuf);  // "<transport xmlns=\"http://www.biigroup.com\">"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY07);
  client.print(sbuf);  // "<header>"
  
  
  
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY08);
  client.print(sbuf);  // "<query id=\""
  client.print(query_id);
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY09);
  client.print(sbuf);  // "\" type=\"storage\">"
  
  free(query_id);

  for(i=0; i<pCount; i++)
  {
	strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY10);
	client.print(sbuf);
	client.print(p[i].pid);
	strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY11);
	client.print(sbuf);
	if(p[i].select == 0){
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY12);
		client.print(sbuf);
	}
	else if(p[i].select == 1){
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY13);
		client.print(sbuf);
	}
	else{
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY14);
		client.print(sbuf);
		client.print(p[i].pLt);
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY15);
		client.print(sbuf);
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY16);
		client.print(sbuf);
		client.print(p[i].pGt);
		strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY17);
		client.print(sbuf);
		
	}
	strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY18);
	client.print(sbuf);
  }

  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY19);
  client.print(sbuf);  // "</query></header></transport></ns2:queryRQ>"
  strcpy_P(sbuf,IEEE1888FetchClient_Get_HTTPBODY20);
  client.print(sbuf);  // "</soapenv:Body></soapenv:Envelope>"
  client.println();
  
  // parse HTTP response
  count = 0;
  
  Serial.println("request send ok"); 

  ieee1888_value* value = NULL;
  while (client.connected()) {

    if (client.available()) {
      c = client.read();  
      if (count == 1 && (c >= '0' && c <= '9')) {  // parse HTTP response code
        rescode = rescode * 10 + (c - '0');
        continue;
      }
      if (c == ' ') {  
        count++;
      }
      if (count == 2 || c == '\n') {  // end of HTTP response code
        break;    
      }
    }
  }

  if (!client.connected()) {  // unexpected disconnect 
	Serial.println("unexpected disconnect");
    client.stop();
	if(value)
	{
		_fetch_deal(value);
		destory(value);
	}
    return IEEE1888_DOWNLOAD_HTTPERR;
  }	
  // disconnect HTTP

  if(client.connected()){
	ret_code = parse(&client,&value);
  }
  
  Serial.println("read remain data");
  while (client.connected() && client.available()) {    
    c = client.read(); // Serial.print(c);  
  }
  client.stop();
  if (rescode == 200) {
	Serial.println("RESPOSE OK!!!");
	if(value)
	{
		_fetch_deal(value);
		destory(value);
	}
    return ret_code;
  }
  return IEEE1888_DOWNLOAD_HTTPERR;
}
	
void IEEE1888FetchClient::ieee1888_mk_new_uuid(char* new_uuid) {
	// generate random uuid
	//char new_uuid[38];
	sprintf(new_uuid, "%04x%04x-%04x-%04x-%04x-%04x%04x%04x", 1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))), 1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))),
			1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))), (1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0)))) | 0x4000,
			(1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0)))) | 0x8000, 1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))),
			1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))), 1 + (int) (((double) 0xffff) * (rand() / (RAND_MAX + 1.0))));

	//return ieee1888_mk_string(new_uuid);
}

PROGMEM prog_char IEEE1888FetchClient_Parse1[] = "point ";
int IEEE1888FetchClient::parse(Client* client, ieee1888_value** value){

   while(client->connected() && client->available()){
      char c=client->read();
      if(c=='<')
	  {
        int i;
        char buf[7];
        for(i=0;i<6 && client->connected() && client->available();i++){
          buf[i]=client->read(); 
		
          if(buf[i]=='>'){ 
			break;
			} 
        }
        buf[i]=NULL_CHARACTER;
		
   	    char sbuf[7];
        strcpy_P(sbuf,IEEE1888FetchClient_Parse1); // "point "	
        if(strcmp(buf,sbuf)==0)
		{
			int s = parseValue(client, value);
			if(s == 0)
				continue;
			else
				return IEEE1888_DOWNLOAD_HTTPERR;
		}
      }
  }

  return IEEE1888_DOWNLOAD_OK;
}

//ieee1888_value* IEEE1888FetchClient::parseValue(Client* client){
//0：成功，1：失败
int IEEE1888FetchClient::parseValue(Client* client, ieee1888_value** value){
 
  int num_1 = 0;  //引号的个数
  int num_2 = 0;  // >的个数
  
  char id[64];
  char time[30];
  char content[10];
  
  while(client->connected() && client->available()){
    char c=client->read(); // Serial.print(c);	
	if(c == '"'){
		num_1++;
		if(num_1 == 1){//第一个引号出现，后面的内容为Pointid
			
			int k = 0;
			while((c = client->read()) != '"')
			{
				id[k++] = c;
			}
			id[k] = '\0';
			num_1++; //第二个引号出现，pointid读取完毕
		
		}
		if(num_1 ==3){//第三个引号出现，后面的内容为time
			
			int l = 0;
			while((c = client->read()) != '"')
			{
				time[l++] = c;
			}
			time[l] = '\0';
			num_1++; //第四个引号出现，time读取完毕
		}
	}
	else if(c == '>')
		num_2++;
	if(num_2 == 2){//第二个>出现，后面的内容为value
		
		int m = 0;
		while((c = client->read()) != '<')
		{
			content[m++] = c;
		}
		content[m] = '\0';
		num_2++;
	}	
	if(num_2 == 4)
	{
		ieee1888_value* pInsert;
		pInsert = (ieee1888_value*)malloc(sizeof(ieee1888_value));
		if(pInsert == NULL){
			Serial.println("************");
			return 1;
		}
		memset(pInsert,0,sizeof(ieee1888_value));
		strcpy(pInsert->pointid,id);
		strcpy(pInsert->time,time);
		strcpy(pInsert->content,content);
		pInsert->next = *value;
		*value = pInsert;
		return 0;		
	}
  }   
  return 1;
}


void IEEE1888FetchClient::destory(ieee1888_value* value){
	ieee1888_value* pNext;
	if(value == NULL){
		return;
	}
	while(value->next != NULL){
		pNext = value->next;
		free(value);
		value = pNext;
	}	
	free(value);
	return;
}