// IEEE1888 Downloader Library for Arduino
//
//  2012/03/07 ver.1   H.Ochiai
//

// --------- IEEE1888FetchClient.h (begin) ---------
#ifndef __IEEE1888FetchClient__
#define __IEEE1888FetchClient__

#include <Client.h>

// return code of get method
#define IEEE1888_DOWNLOAD_OK       0      // Succeeded
#define IEEE1888_DOWNLOAD_CONNFAIL 1      // Connection faild (Socket I/O error)
#define IEEE1888_DOWNLOAD_DNSERR   2      // DNS error
#define IEEE1888_DOWNLOAD_HTTPERR  3      // HTTP Server error (The response was not "200 OK")
#define IEEE1888_DOWNLOAD_IEEE1888ERR  4      // IEEE1888 Server error
#define IEEE1888_DOWNLOAD_INTERNAL_ERR 5  // INTERNAL ERROR (Client side) e.g., memory overflow

// class definition
#define MAX_IEEE1888_POINTID_LEN 64



typedef struct ieee1888_value{
	char time[30];
	char content[10];
	char pointid[64];
	struct ieee1888_value* next;
}ieee1888_value;



typedef struct fetch_point {
  char pid[MAX_IEEE1888_POINTID_LEN];
  int select;
  char pGt[32];
  char pLt[32];
}fetch_point;


typedef int (*fetch_deal)(ieee1888_value* fetch);

class IEEE1888FetchClient {
public: 
  void begin(const uint8_t* server_ip4,
             const char* server_host,
             const char* server_path,
             uint16_t server_port,
			 fetch_deal _fetch_deal);
             
		 
  int get(fetch_point p[],int pCount);		 
  
  void ieee1888_mk_new_uuid(char*);
  int parse(Client* client, ieee1888_value** value);
  int parseValue(Client* client, ieee1888_value** value);
  void destory(ieee1888_value* value);
    fetch_deal  _fetch_deal;
private: 
  const uint8_t* server_ip4;
  const char* server_host;
  const char* server_path;
  uint16_t server_port;
  

};

#endif  // #ifndef IEEE1888FetchClient
