#include <SPI.h>       
#include <Ethernet.h>        
#include <EthernetClient.h>  
#include <EthernetServer.h>        
#include <avr/pgmspace.h>
#include "TinyXML.h"
#include "IEEE1888ServerAgent.h"
#include <EEPROM.h>

//File myFile;

query_handler _query = NULL;
data_handler _data = NULL;
fault_handler _fault = NULL;
void wirteSuccessful(EthernetClient client);
void writeFailed(EthernetClient client,char* errorMsg);
//void fetchSuccessful(EthernetClient client,char* value);
void fetchSuccessful(EthernetClient client,char* time,char* value);
void fetchFailed(EthernetClient client);

void setHandler(
	query_handler query,
	data_handler data, 
	fault_handler fault){
		_query= query;
		_data= data;
		_fault=fault;
}

TinyXML xml;
uint8_t buffer[150];
uint16_t buflen = 150;
typedef struct point {
	char pointID[64];
	char value[16];
	char method[10];
	char time[30];
}point;
point _point;

PROGMEM prog_char DATA_TAGNAME[] =  "/soapenv:Envelope/soapenv:Body/ns2:dataRQ/transport/body/point";
PROGMEM prog_char DATA_VALUE[] =  "/soapenv:Envelope/soapenv:Body/ns2:dataRQ/transport/body/point/value";
PROGMEM prog_char QUERY_TAGNAME[] =  "/soapenv:Envelope/soapenv:Body/ns2:queryRQ/transport/header/query";
PROGMEM prog_char QUERY_KEY_TAGNAME[] = "/soapenv:Envelope/soapenv:Body/ns2:queryRQ/transport/header/query/key";
PROGMEM prog_char QUERY_REQ[] = "/soapenv:Envelope/soapenv:Body/ns2:queryRQ";
PROGMEM prog_char DATA_REQ[] = "/soapenv:Envelope/soapenv:Body/ns2:dataRQ";

int isPlaceID= 0;
int queryCopy = 0;
char xmlBuff[80];
void XML_callback( uint8_t statusflags, char* tagName,  uint16_t tagNameLen,  char* data,  uint16_t dataLen )
{

	memset(xmlBuff,0,80);
	if (statusflags & STATUS_START_TAG)
	{
		if ( tagNameLen )
		{
			//Serial.print("Start tag ");
			//Serial.println(tagName);
			strcpy_P(xmlBuff,QUERY_TAGNAME);
			if (strcmp(tagName,xmlBuff)==0){
				queryCopy=1;
			}
		}
	}
	else if  (statusflags & STATUS_END_TAG)
	{
		//Serial.print("End tag ");
		//Serial.println(tagName);
		strcpy_P(xmlBuff,QUERY_KEY_TAGNAME);
		if (strcmp(tagName,xmlBuff)==0){
				queryCopy=0;
		}
	}else if  (statusflags & STATUS_TAG_TEXT)
	{
//	Serial.println(tagName);
		strcpy_P(xmlBuff,DATA_TAGNAME);

		if (strcmp(tagName,xmlBuff) == 0){
			strcpy(_point.pointID,data);
		}
		
		memset(xmlBuff,0,80);
		strcpy_P(xmlBuff,DATA_VALUE);
		if (strcmp(tagName,xmlBuff) == 0){
			strcpy(_point.value,data);
		}
		
		memset(xmlBuff,0,80);
		strcpy_P(xmlBuff,QUERY_REQ);
		if (strcmp(tagName,xmlBuff) == 0){
			strcpy(_point.method,"query");
		}
		
		memset(xmlBuff,0,80);
		strcpy_P(xmlBuff,DATA_REQ);
		if (strcmp(tagName,xmlBuff) == 0){
			strcpy(_point.method,"data");
		}
		
		memset(xmlBuff,0,80);
		strcpy_P(xmlBuff,QUERY_TAGNAME);
		if (strcmp(tagName,xmlBuff) == 0){
			isPlaceID = 1;
		}
		
	}
	else if  (statusflags & STATUS_ATTR_TEXT)
	{
		if(strcmp(tagName,"id")==0 && (isPlaceID == 1)){
			strcpy(_point.pointID,data);
		}
		if (strcmp(tagName,"time") == 0){
			strcpy(_point.time,data);
		}
	}
	else if  (statusflags & STATUS_ERROR){}
}

EthernetServer server(1888);

void serverCreate(){
	xml.init((uint8_t*)&buffer,buflen,&XML_callback);
	server.begin();
	
/*	if (!SD.begin(4)) {  //如果从CS口与SD卡通信失败，串口输出信息initialization failed!
		Serial.println("initialization failed!");
		return;
	}
	Serial.println("initialization done.");  //CS口与SD卡通信成功，串口输出信息initialization done.
 
	if (SD.exists("example.txt")) {  //检查example.txt文件是否存在
		Serial.println("example.txt exists.");  //如果存在输出信息example.txt exists.至串口
	}
	else {
		Serial.println("example.txt doesn't exist.");  //不存在输出信息example.txt doesn't exist.至串口
	}
 
	Serial.println("Creating example.txt...");
	myFile = SD.open("example.txt", FILE_WRITE);  //打开example.txt，如果没有自动创建，写入状态
	myFile.close();  //关闭文件
 
	if (SD.exists("example.txt")) {
		Serial.println("example.txt exists.");
	}
	else {
		Serial.println("example.txt doesn't exist.");  
	}*/
}

char queryPre[200];

void serverRun(){

	EthernetClient client = server.available();

	if (client) {	
		//while (client.connected()){	
			Serial.println("connect");
			int index = 0;	
			while (client.connected())
			{
				if (client.available() > 0)
				{
					char c = client.read();
					//Serial.print(c);
					if (queryCopy) {
						queryPre[index] = c;
						index++;
					}					
					xml.processChar(c);
				}else{
					delay(200);//important wait for body package.
					if (client.available()){
						continue;
					}else{
						//Serial.println("break");
						break;
					}
				}
			}

			//Serial.println(queryPre);
			
			if (strcmp(_point.method,"data")==0){
				int result = _data(_point.pointID,_point.value,_point.time);
				if (result != 0){					
					writeFailed(client,"server internal error.");
				}else{
					wirteSuccessful(client);
				}
				
			}else if (strcmp(_point.method,"query")==0){
				char ID[64];
				char TIME[30];
				char VALUE[16];
				_query(ID,TIME,VALUE);
			//	Serial.println(ID);
			//	Serial.println(TIME);
			//	Serial.println(VALUE);
				if(strcmp(ID,_point.pointID) == 0){
					fetchSuccessful(client,TIME,VALUE);
				}
				else{
					fetchFailed(client);
				}
				/*char *result = _query(_point.pointID);
				if (result == NULL){
					fetchFailed(client);
				}else{
					fetchSuccessful(client,result);
				}*/
			}else{
				//client.print("request format error \r\n");
				//Serial.println("request format error!!");
			}
			//client.stop();			
			memset(&_point,0,sizeof(_point));
	//	}
	}
	//wait for client get the data
	delay(100);
}

PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER01[] =  "POST ";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER02[] =  "HTTP/1.1";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER03[] =  "Content-Type: text/xml;charset=utf-8";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER04[] =  "User-Agent: IEEE1888ServerAgent (Arduino)";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER05[] =  "Host: ";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER06[] =  "SOAPAction: \"http://soap.ieee1888.org/data\"";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPHEADER07[] =  "Content-Length: ";

// Messages (Stored in the program memory) -- HTTP Body Part
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY01[] =  "<?xml version='1.0' encoding='UTF-8'?>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY02[] =  "<soapenv:Envelope xmlns:soapenv=";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY03[] =  "\"http://schemas.xmlsoap.org/soap/envelope/\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY04[] =  "<soapenv:Body>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY05[] =  "<ns2:dataRS xmlns:ns2=\"http://soap.ieee1888.org/\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY06[] =  "<transport xmlns=\"http://www.biigroup.com\">";

PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY07[] =  "<header>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY08[] =  "</header>";

PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY09[] =  "<body>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY10[] =  "<point id=\"";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY11[] =  "\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY12[] =  "<value time=\"";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY13[] =  "\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY14[] =  "</value>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY15[] =  "</point>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY16[] =  "</body>";

PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY17[] =  "</transport>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY18[] =  "</ns2:dataRS>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY19[] =  "</soapenv:Body>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY20[] =  "</soapenv:Envelope>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY21[] =  "</ns2:queryRS>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY22[] =  "<ns2:queryRS xmlns:ns2=\"http://soap.ieee1888.org/\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY23[] =  "</transport>";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY24[] =  "<error type=\"POINT_NOT_FOUND\">";
PROGMEM prog_char IEEE1888UploadAgent_Post_HTTPBODY25[] =  "point was not found.</error>";
void wirteSuccessful(EthernetClient client){
	char sbuf[100];
	int clen = 303;

	// send HTTP header
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER02);
	client.print(sbuf); // " HTTP/1.1"
	client.print(" 200 OK\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER03);
	client.print(sbuf); // "Content-Type: text/xml; charset=UTF-8"
	client.print("\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER07);
	client.print(sbuf); // "Content-Length: "
	client.print(clen);
	client.print("\r\n\r\n");

	// send HTTP body
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY01);
	client.print(sbuf); // "<?xml version='1.0' encoding='UTF-8'?>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY02);
	client.print(sbuf); // "<soapenv:Envelope xmlns:soapenv="
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY03);
	client.print(sbuf); // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY04);
	client.print(sbuf); // "<soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY05);
	client.print(sbuf); // "<ns2:dataRQ xmlns:ns2=\"http://soap.ieee1888.org/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY06);
	client.print(sbuf); // "<transport xmlns=\"http://www.biigroup.com\">";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY07);
	client.print(sbuf); // "<header>";
	client.print("<OK />");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY08);
	client.print(sbuf); // "</header>";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY17);
	client.print(sbuf); // "</transport>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY18);
	client.print(sbuf); // "</ns2:dataRQ>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY19);
	client.print(sbuf); // "</soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY20);
	client.print(sbuf); // "</soapenv:Envelope>"
	client.print("\r\n");

}

void writeFailed(EthernetClient client,char* errorMsg){
	char sbuf[50];
	int clen = 303;

	// send HTTP header
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER02);
	client.print(sbuf); // " HTTP/1.1"
	client.print(" 200 OK\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER03);
	client.print(sbuf); // "Content-Type: text/xml; charset=UTF-8"
	client.print("\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER07);
	client.print(sbuf); // "Content-Length: "
	client.print(clen);
	client.print("\r\n\r\n");

	// send HTTP body
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY01);
	client.print(sbuf); // "<?xml version='1.0' encoding='UTF-8'?>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY02);
	client.print(sbuf); // "<soapenv:Envelope xmlns:soapenv="
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY03);
	client.print(sbuf); // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY04);
	client.print(sbuf); // "<soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY05);
	client.print(sbuf); // "<ns2:dataRQ xmlns:ns2=\"http://soap.ieee1888.org/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY06);
	client.print(sbuf); // "<transport xmlns=\"http://www.biigroup.com\">";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY07);
	client.print(sbuf); // "<header>";
	client.print("<error type=\"request\">");
	client.print(errorMsg);
	client.print("</error>");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY08);
	client.print(sbuf); // "</header>";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY17);
	client.print(sbuf); // "</transport>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY18);
	client.print(sbuf); // "</ns2:dataRQ>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY19);
	client.print(sbuf); // "</soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY20);
	client.print(sbuf); // "</soapenv:Envelope>"
	client.print("\r\n");
}

//void fetchSuccessful(EthernetClient client,char* value){
  void fetchSuccessful(EthernetClient client,char* time,char* value){
	char sbuf[50];
	int clen = 404;
	
	Serial.println("fetchsuccessful");

	// send HTTP header
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER02);
	client.print(sbuf); // " HTTP/1.1"
	client.print(" 200 OK\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER03);
	client.print(sbuf); // "Content-Type: text/xml; charset=UTF-8"
	client.print("\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER07);
	client.print(sbuf); // "Content-Length: "
	client.print(clen);
	client.print("\r\n\r\n");

	// send HTTP body
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY01);
	client.print(sbuf); // "<?xml version='1.0' encoding='UTF-8'?>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY02);
	client.print(sbuf); // "<soapenv:Envelope xmlns:soapenv="
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY03);
	client.print(sbuf); // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY04);
	client.print(sbuf); // "<soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY22);
	client.print(sbuf); // "<ns2:queryRS xmlns:ns2=\"http://soap.ieee1888.org/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY06);
	client.print(sbuf); // "<transport xmlns=\"http://www.biigroup.com\">";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY07);
	client.print(sbuf); // "<header>";
	client.print("<OK />");
	client.print("<query ");
	client.print(queryPre);
	client.print("</query>");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY08);
	client.print(sbuf); // "</header>";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY09);
	client.print(sbuf);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY10);
	client.print(sbuf);
	client.print(_point.pointID);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY11);
	client.print(sbuf);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY12);
	client.print(sbuf);
	//client.print("2013-11-04T13:21:00.000+08:00");
	client.print(time);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY13);
	client.print(sbuf);
	client.print(value);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY14);
	client.print(sbuf);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY15);
	client.print(sbuf);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY16);
	client.print(sbuf);

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY17);
	client.print(sbuf); // "</transport>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY21);
	client.print(sbuf); // "</ns2:queryRS>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY19);
	client.print(sbuf); // "</soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY20);
	client.print(sbuf); // "</soapenv:Envelope>"
	client.print("\r\n");
}

void fetchFailed(EthernetClient client){
	char sbuf[50];
	int clen = 404;

	// send HTTP header
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER02);
	client.print(sbuf); // " HTTP/1.1"
	client.print(" 200 OK\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER03);
	client.print(sbuf); // "Content-Type: text/xml; charset=UTF-8"
	client.print("\r\n");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPHEADER07);
	client.print(sbuf); // "Content-Length: "
	client.print(clen);
	client.print("\r\n\r\n");

	// send HTTP body
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY01);
	client.print(sbuf); // "<?xml version='1.0' encoding='UTF-8'?>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY02);
	client.print(sbuf); // "<soapenv:Envelope xmlns:soapenv="
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY03);
	client.print(sbuf); // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY04);
	client.print(sbuf); // "<soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY22);
	client.print(sbuf); // "<ns2:queryRS xmlns:ns2=\"http://soap.ieee1888.org/\">"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY06);
	client.print(sbuf); // "<transport xmlns=\"http://www.biigroup.com\">";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY07);
	client.print(sbuf); // "<header>";
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY24);
	client.print(sbuf);
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY25);
	client.print(sbuf);
	client.print("<query ");
	client.print(queryPre);
	client.print("</query>");
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY08);
	client.print(sbuf); // "</header>";

	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY17);
	client.print(sbuf); // "</transport>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY21);
	client.print(sbuf); // "</ns2:queryRS>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY19);
	client.print(sbuf); // "</soapenv:Body>"
	strcpy_P(sbuf,IEEE1888UploadAgent_Post_HTTPBODY20);
	client.print(sbuf); // "</soapenv:Envelope>"
	client.print("\r\n");
}

void fetchData(char* pointID,char* time,char* value)
{
	int i;
	if(EEPROM.read(0)!=0xaa || EEPROM.read(1)!=0x55 || EEPROM.read(2) != 0x66) 
	{
		return ;
	} 
	for(i=0; i<64; i++){
		pointID[i] = EEPROM.read(i+3);
	} 
	for(i=0; i<30; i++){
		time[i] = EEPROM.read(i+67);
	} 
	for(i=0; i<16; i++){
		value[i] = EEPROM.read(i+97);
	} 
}

void writeData(char* pointID,char* time,char* value){
   int i = 0;
   EEPROM.write(0,0xaa);
   EEPROM.write(1,0x55);
   EEPROM.write(2,0x66);
	for(i=0; i<64; i++){
		EEPROM.write(i+3,pointID[i]);
	} 
	for(i=0; i<30; i++){
		EEPROM.write(i+67,time[i]);
	} 
	for(i=0; i<16; i++){
		EEPROM.write(i+97,value[i]);
	} 
}