//  IEEE1888 / IEEE1888 Uploader Library for Arduino

// --------- IEEE1888WriteClient.cpp (begin) ---------
#include <SPI.h>              
#include <Ethernet.h>        
#include <EthernetClient.h>           
#include <avr/pgmspace.h>         
#include "IEEE1888WriteClient.h"

// void void IEEE1888WriteClient::begin( ... );
void IEEE1888WriteClient::begin(
		  const char* server_host,
		  const char* server_path,
		  unsigned short server_port )
{
	this->server_host=server_host;
	this->server_path=server_path;
	this->server_port=server_port;
}

// Messages (Stored in the program memory) -- HTTP 头
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER01[] =  "POST ";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER02[] =  " HTTP/1.1";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER03[] =  "Content-Type: text/xml charset=UTF-8";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER04[] =  "User-Agent: IEEE1888WriteClient (Arduino HCU)";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER05[] =  "Host: ";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER06[] =  "SOAPAction: \"http://soap.ieee1888.org/data\"";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPHEADER07[] =  "Content-Length: ";

// Messages (Stored in the program memory) -- HTTP 实体
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY01[] =  "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY02[] =  "<soapenv:Envelope xmlns:soapenv=";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY03[] =  "\"http://schemas.xmlsoap.org/soap/envelope/\">";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY04[] =  "<soapenv:Body>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY05[] =  "<ns2:dataRQ xmlns:ns2=\"http://soap.ieee1888.org/\">";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY06[] =  "<transport xmlns=\"http://www.biigroup.com\">";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY07[] =  "<body>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY10[] =  "<point id=\"";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY11[] =  "\">";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY12[] =  "<value time=\"";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY13[] =  "\">";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY14[] =  "</value>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY15[] =  "</point>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY17[] =  "</body>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY18[] =  "</transport>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY19[] =  "</ns2:dataRQ>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY20[] =  "</soapenv:Body>";
PROGMEM prog_char IEEE1888WriteClient_Post_HTTPBODY21[] =  "</soapenv:Envelope>";

// Messages (Stored in the program memory) -- 时间序列格式
PROGMEM prog_char IEEE1888WriteClient_Post_TimeFormat[] = "%04d-%02d-%02dT%02d:%02d:%02d%s\0";

/*
 int IEEE1888WriteClient::post(struct ieee1888_element* v, byte esize);
 @param:
 		struct ieee1888_element* v:上报数据的指针
 		esize 					  ：上报个数
 @return values:
        #define IEEE1888_UPLOAD_OK       0  
        #define IEEE1888_UPLOAD_CONNFAIL 1  
        #define IEEE1888_UPLOAD_DNSERR   2  
        #define IEEE1888_UPLOAD_HTTPERR  3  
		#define IEEE1888_UPLOAD_IEEE1888ERR 4
        #define IEEE1888_UPLOAD_FATAL   99  
*/
int IEEE1888WriteClient::post(struct ieee1888_element* v, byte esize){

  int rescode = 0;  
  int clen = 0;     
  struct ieee1888_element *v0;
  char count;
  unsigned char c;
  char sbuf[55];

  EthernetClient client;

  if (!client.connect(server_host, server_port)) 
  {
    return(IEEE1888_UPLOAD_CONNFAIL); 
  }
  v0 = v;
  
  /*  注明：在HTTP请求中Content-Length指明指明实体正文的长度，包括其中的回车换行符  */  

  /* strlen(IEEE1888WriteClient_Post_HTTPBODY01)+...strlen(IEEE1888WriteClient_Post_HTTPBODY07)+
	 strlen(IEEE1888WriteClient_Post_HTTPBODY17)+...strlen(IEEE1888WriteClient_Post_HTTPBODY21)+ 2*12(12指12个回车换行符个数)
	 由于arduino内存空间较少，并且这些内容基本固定故此处写死，用户在实际使用中可对上述值的修改相应改变该值 
  */
	clen = 317;
	
	for (count = 0; count < esize; count++) {
		clen += strlen(v0->cid);
		clen += strlen(v0->value);
		
		/*strlen(IEEE1888WriteClient_Post_HTTPBODY10)+....+strlen(IEEE1888WriteClient_Post_HTTPBODY15)+
		  strlen(IEEE1888WriteClient_Post_TimeFormat) + 2*3(3指回车换行符个数)*/
		clen += 75;
		
		v0++;
  }

  // 发送HTTP头
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER01);
  client.print(sbuf);  // "POST "
  client.print(server_path); 
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER02);
  client.println(sbuf); // " HTTP/1.1"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER03);
  client.println(sbuf); // "Content-Type: text/xml; charset=UTF-8"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER04);
  client.println(sbuf); // "User-Agent: IEEE1888WriteClient (Arduino HCU)"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER05);
  client.print(sbuf); // "Host: "
  client.println(server_host);
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER06);
  client.println(sbuf); // "SOAPAction: \"http://soap.ieee1888.org/data\""
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPHEADER07);
  client.print(sbuf); // "Content-Length: "
  client.println(clen);
  client.println();

  // 发送HTTP实体
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY01);
  client.println(sbuf); // "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY02);
  client.print(sbuf); // "<soapenv:Envelope xmlns:soapenv="
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY03);
  client.println(sbuf); // "\"http://schemas.xmlsoap.org/soap/envelope/\">"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY04);
  client.println(sbuf); // "<soapenv:Body>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY05);
  client.println(sbuf); // "<ns2:dataRQ xmlns:ns2=\"http://soap.ieee1888.org/\">"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY06);
  client.println(sbuf); // "<transport xmlns=\"http://ieee1888.org/ieee1888/2009/11/\">"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY07);
  client.println(sbuf); // "<body>"

  v0=v;
  for (count = 0; count < esize; count++) {
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY10);
    client.print(sbuf); // "<point id=\""
	client.print(v0->cid); 
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY11);
    client.println(sbuf); // "\">"   
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY12);
    client.print(sbuf); // "<value time=\""
    client.print(element_time_to_str(v0)); 
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY13);
    client.print(sbuf); // "\">"
    client.print(v0->value);   
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY14);
    client.println(sbuf); // "</value>"
    strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY15);
    client.println(sbuf); // "</point>"
    v0++;
  }

  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY17);
  client.println(sbuf); // "</body>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY18);
  client.println(sbuf); // "</transport>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY19);
  client.println(sbuf); // "</ns2:dataRQ>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY20);
  client.println(sbuf); // "</soapenv:Body>"
  strcpy_P(sbuf,IEEE1888WriteClient_Post_HTTPBODY21);
  client.println(sbuf); // "</soapenv:Envelope>"
  client.println();

  // 解析HTTP响应
  count = 0;
  while (client.connected()) {

    if (client.available()) {
      c = client.read();  
      if (count == 1 && (c >= '0' && c <= '9')) {  
        rescode = rescode * 10 + (c - '0');
        continue;
      }
      if (c == ' ') {  
        count++;
      }
      if (count == 2 || c == '\n') {  
        break;    
      }
    }
  }
  if (!client.connected()) {  
    client.stop();
    return(IEEE1888_UPLOAD_FATAL);
  }

  // HTTP连接终止
  while (client.connected() && client.available()) {
    c = client.read(); 
  }
  client.stop();
  if (rescode == 200) {
    return(IEEE1888_UPLOAD_OK);
  }
  return(IEEE1888_UPLOAD_HTTPERR);
}


char* IEEE1888WriteClient::element_time_to_str(struct ieee1888_element* e)
{ 
  static char ieee1888_time[28]; 
  char str_timeformat[35];
  strcpy_P(str_timeformat,IEEE1888WriteClient_Post_TimeFormat);
  sprintf(ieee1888_time,str_timeformat,e->year,e->month,e->day,e->hour,e->minute,e->second,e->timezone);
  return ieee1888_time;
}
// --------- IEEE1888WriteClient.cpp (end) ---------
