
#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h>
#include <IEEE1888WriteClient.h>
#include <Time.h>
#include <Dhcp.h>
#include <Dns.h>
#include <EEPROM.h>
#include <MsTimer2.h>
#include <avr/wdt.h>

#define MAX_SERVER_HOST_LEN 64  //服务端主机名最大长度
#define MAX_SERVER_PATH_LEN 64  //服务端存储路径最大长度

#define FIRMWARE_VERSION 0x01       

unsigned int MEASURE_INTERVAL = 60;  
char timezone[] = "+08:00";           
const long timezoneOffset = +28800L; 


byte local_mac[6]={0xb0,0x12,0x66,0x06,0x01,0x41};   
IPAddress local_ip4(10, 1, 1, 188);  //
IPAddress local_gateway(10, 1, 1, 1);
IPAddress local_netmask(255, 255, 255, 0 );
IPAddress dns_addr(8,8,8,8) ;
boolean dhcp_e = true;

char server_host[MAX_SERVER_HOST_LEN] = "10.1.1.202";       //默认服务端主机名(IP地址)
char server_path[MAX_SERVER_PATH_LEN] = "/IEEEstorage";     //默认服务端存储路径    
int server_port = 1888;                                     //默认服务端端口号

static byte ntp_server1[4] = { 10,1,1,131 };   // 默认NTP服务端地址
static byte ntp_server2[4] = { 202,120,2,101};     

unsigned int voltage_setting=10;

struct ieee1888_element element[]=
{
  {"http://biigroup.cn/2508/Humidity", "100.0",0,0,0,0,0,0,timezone},
  {"http://biigroup.cn/2501/Illumination", "500.0",0,0,0,0,0,0,timezone},
};

static byte config_mode=0;     // 标志是否配置模式
static char inLine[135];       // 读取串口输入内容
static byte nInLine=0;       
long serial_check_count = 0;  

IEEE1888WriteClient IEEE1888;
EthernetUDP udp;
static char ieee1888upload_status=-1;

PROGMEM prog_char bootProcessHeader1[] = "Starting IEEE1888 writeclient demo Ver.1.0 ... ";
int COUNT = sizeof(element)/sizeof(ieee1888_element);

void setup()
{
  int i;
  char outLine[60];
  Serial.begin(9600);
  strcpy_P(outLine,bootProcessHeader1);
  Serial.println(outLine);
  saveConfig(); //保存程序中的配置信息
  loadConfig(); //读取配置信息

  /*初始化网口*/
  Serial.print("local_ip=");
  Serial.println(local_ip4);
  Serial.print("local_gateway=");
  Serial.println(local_gateway);
  Serial.print("local_netmask=");
  Serial.println(local_netmask);
  
  Ethernet.begin(local_mac, local_ip4, dns_addr, local_gateway, local_netmask);
  
  Serial.print("server_host=");
  Serial.println(server_host);
  Serial.print("server_path=");
  Serial.println(server_path);
  Serial.print("server_port=");
  Serial.println(server_port);

  IEEE1888.begin(server_host, server_path, server_port);
  setSyncProvider(getNtpTime);
  Serial.println("init ok!!!!");
	 
  if(config_mode){
    Serial.print("# ");    //配置模式中
  }else{
    Serial.print("> ");    //运行模式中
  }
  
}

byte error_count=0;

void loop()
{
	int i;
	time_t t;

    parseSerial(COUNT);
     
    t = now();
    for(i = 0; i < COUNT; i++) {
         element[i].year = year(t);
         element[i].month = month(t);
         element[i].day = day(t);
         element[i].hour = hour(t);
         element[i].minute = minute(t);
         element[i].second = second(t);
      }
	ieee1888upload_status=IEEE1888.post(element, COUNT);

    if(ieee1888upload_status==IEEE1888_UPLOAD_OK)
	{
        error_count=0;	
    }else
	{
		error_count++;
    }
 }  

PROGMEM prog_char parseSerial_Error1[] = "Exceeded maxium length ... force new line";
PROGMEM prog_char parseSerial_Error2[] = "Schema Error: ";
PROGMEM prog_char parseSerial_Error3[] = "First, type 'conf' to get into configuration mode.";
PROGMEM prog_char parseSerial_Error4[] = "Already in configuration mode.\0";
PROGMEM prog_char parseSerial_Error5[] = "Cannot exit from the configuration mode.\0";
PROGMEM prog_char parseSerial_Error6[] = "Already in running mode.\0";
PROGMEM prog_char parseSerial_Error7[] = "Syntax Error: ";
PROGMEM prog_char parseSerial_Error8[] = "Readonly parameter.";
PROGMEM prog_char parseSerial_Error9[] = "Too Long: ";

PROGMEM prog_char parseSerial_Indent[] = "   " ;
PROGMEM prog_char parseSerial_Is_OK[] = " ... OK" ;
PROGMEM prog_char parseSerial_MAC_Is_Format[] = "MAC=%02x:%02x:%02x:%02x:%02x:%02x";
PROGMEM prog_char parseSerial_IP_Is_Format[] = "IP=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_NM_Is_Format[] = "NM=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_GW_Is_Format[] = "GW=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_DNS_Is_Format[] = "DNS=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_NTP1_Is_Format[] = "NTP1=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_NTP2_Is_Format[] = "NTP2=%d.%d.%d.%d";
PROGMEM prog_char parseSerial_STAT_NONE_Format[] = "STAT=NONE";
PROGMEM prog_char parseSerial_STAT_OK_Format[] = "STAT=OK";
PROGMEM prog_char parseSerial_STAT_NG_Format[] = "STAT=NG(%s)";

PROGMEM prog_char parseSerial_show_Message00[] = "--- Local Network Configuration ---";
PROGMEM prog_char parseSerial_show_Message01[] = "--- IEEE1888 Configuration ---";
PROGMEM prog_char parseSerial_show_Message02[] = "--- NTP Server Configuration ---";
PROGMEM prog_char parseSerial_show_Message03[] = "AutogenID(";
PROGMEM prog_char parseSerial_show_Message04[] = "): ";
PROGMEM prog_char parseSerial_show_Message05[] = "--- Upload Status ---";

PROGMEM prog_char parseSerial_help_Message00[] = "IEEE1888 Power Meter ver.1.0";
PROGMEM prog_char parseSerial_help_Message01[] = "";
PROGMEM prog_char parseSerial_help_Message02[] = "-- Basic Command Set --";
PROGMEM prog_char parseSerial_help_Message03[] = "conf: goes into configuration mode (indicated as '# ')";
PROGMEM prog_char parseSerial_help_Message04[] = "exit: returns to running mode (indicated as '> ')";
PROGMEM prog_char parseSerial_help_Message05[] = "show: prints out the config";
PROGMEM prog_char parseSerial_help_Message06[] = "save: saves the config into the EEPROM";
PROGMEM prog_char parseSerial_help_Message07[] = "help or ?: prints out this message";
PROGMEM prog_char parseSerial_help_Message08[] = "";
PROGMEM prog_char parseSerial_help_Message09[] = "-- Configuration Parameters --";
PROGMEM prog_char parseSerial_help_Message10[] = "* Local Network";
PROGMEM prog_char parseSerial_help_Message11[] = " - MAC=xx:xx:xx:xx:xx:xx (readonly)";
PROGMEM prog_char parseSerial_help_Message12[] = " - DHCP={true, false}";
PROGMEM prog_char parseSerial_help_Message13[] = " - IP=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message14[] = " - NM=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message15[] = " - GW=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message16[] = " - DNS=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message17[] = "* IEEE1888 (Server & PointSet ID & Value)";
PROGMEM prog_char parseSerial_help_Message18[] = " - HOST=....";
PROGMEM prog_char parseSerial_help_Message19[] = " - PATH=....";
PROGMEM prog_char parseSerial_help_Message20[] = " - PORT=....";
PROGMEM prog_char parseSerial_help_Message21[] = " - ID=http://....";
PROGMEM prog_char parseSerial_help_Message22[] = " - INT=...";
PROGMEM prog_char parseSerial_help_Message23[] = " - V=...";
PROGMEM prog_char parseSerial_help_Message24[] = "* NTP Time Server";      
PROGMEM prog_char parseSerial_help_Message25[] = " - NTP1=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message26[] = " - NTP2=x.x.x.x";
PROGMEM prog_char parseSerial_help_Message27[] = "* Status";
PROGMEM prog_char parseSerial_help_Message28[] = " - STAT=.... (readonly)";
PROGMEM prog_char parseSerial_help_Message29[] = "";

PROGMEM const char* parseSerial_help_Message_List[]={
  parseSerial_help_Message00,  parseSerial_help_Message01,  parseSerial_help_Message02,  parseSerial_help_Message03,
  parseSerial_help_Message04,  parseSerial_help_Message05,  parseSerial_help_Message06,  parseSerial_help_Message07,
  parseSerial_help_Message08,  parseSerial_help_Message09,  parseSerial_help_Message10,  parseSerial_help_Message11,
  parseSerial_help_Message12,  parseSerial_help_Message13,  parseSerial_help_Message14,  parseSerial_help_Message15,
  parseSerial_help_Message16,  parseSerial_help_Message17,  parseSerial_help_Message18,  parseSerial_help_Message19,
  parseSerial_help_Message20,  parseSerial_help_Message21,  parseSerial_help_Message22,  parseSerial_help_Message23,
  parseSerial_help_Message24,  parseSerial_help_Message25,  parseSerial_help_Message26,  parseSerial_help_Message27,
  parseSerial_help_Message28,  parseSerial_help_Message29
};

/* 解析串口信息完成参数配置，count为发送数据的个数 */
void parseSerial(int count){
  
  byte is_newline=0;
  char outLine[60];

  serial_check_count++;
  while(Serial.available()){

    serial_check_count=0;
    if(nInLine<134){
      char c=Serial.read();
      
      if(c==0x7f){
        // Back Space (BS)
        if(nInLine>0){
          Serial.print(c);
          inLine[--nInLine]='\0';
        }
      }else if(0x20<=c && c<=0x7e){
        Serial.print(c);
        inLine[nInLine++]=c;
      }else if(c==0x0d){
        is_newline=1;
        inLine[nInLine]=0x00;
        Serial.println();
        break; 
      }
    }else{
      Serial.println();
      strcpy_P(outLine,parseSerial_Error1);
      Serial.println(outLine); 
      nInLine--;
      is_newline=1;
      inLine[nInLine]=0x00;
      break;
    }
  }

  if(is_newline){

    /*解析MAC地址*/
     if(inLine[0]=='M' && inLine[1]=='A' && inLine[2]=='C' && inLine[3]=='='){
	   
	   if(config_mode){
       byte mac[6];
       if(parseMACExpr(&inLine[4],mac)){
         char tmp[30];
         // PROGMEM prog_char parseSerial_MAC_Is_Format[] = "MAC=%d.%d.%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_MAC_Is_Format);
         sprintf(tmp,outLine,mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
         local_mac[0]=mac[0]; local_mac[1]=mac[1]; local_mac[2]=mac[2]; local_mac[3]=mac[3];local_mac[4]=mac[4];local_mac[5]=mac[5];
         Serial.print(tmp);
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
    /* 解析IP */
    }else if(inLine[0]=='I' && inLine[1]=='P' && inLine[2]=='='){     
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[3],ip)){
         char tmp[20];
         // PROGMEM prog_char parseSerial_IP_Is_Format[] = "IP=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_IP_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         local_ip4[0]=ip[0]; local_ip4[1]=ip[1]; local_ip4[2]=ip[2]; local_ip4[3]=ip[3];
         Serial.print(tmp);
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
    /* 解析网关 */
    }else if(inLine[0]=='G' && inLine[1]=='W' && inLine[2]=='='){
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[3],ip)){
         char tmp[20];
         // PROGMEM prog_char parseSerial_GW_Is_Format[] = "GW=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_GW_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         local_gateway[0]=ip[0]; local_gateway[1]=ip[1]; local_gateway[2]=ip[2]; local_gateway[3]=ip[3];
         Serial.print(tmp); 
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }     
    /*   解析netmask  */
    }else if(inLine[0]=='N' && inLine[1]=='M' && inLine[2]=='='){
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[3],ip)){
         char tmp[20];
         // PROGMEM prog_char parseSerial_NM_Is_Format[] = "NM=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_NM_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         local_netmask[0]=ip[0]; local_netmask[1]=ip[1]; local_netmask[2]=ip[2]; local_netmask[3]=ip[3];
         Serial.print(tmp); 
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }      
    /* 解析DNS */
    }else if(inLine[0]=='D' && inLine[1]=='N' && inLine[2]=='S' && inLine[3]=='='){
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[4],ip)){
         char tmp[20];
         // PROGMEM prog_char parseSerial_DNS_Is_Format[] = "DNS=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_DNS_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         dns_addr[0]=ip[0]; dns_addr[1]=ip[1]; dns_addr[2]=ip[2]; dns_addr[3]=ip[3];
         Serial.print(tmp); 
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
    /* 解析主机 */    
    }else if(inLine[0]=='H' && inLine[1]=='O' && inLine[2]=='S' && inLine[3]=='T' && inLine[4]=='='){
      if(config_mode){
		/*按服务名解析*/
        if( ('A'<=inLine[5] && inLine[5]<='Z') ||
            ('a'<=inLine[5] && inLine[5]<='z') ||
            inLine[5]=='_' ){        
          if(strlen(&inLine[5])<MAX_SERVER_HOST_LEN){
            strcpy(server_host,&inLine[5]);
            Serial.print("HOST="); Serial.print(server_host); 
            strcpy_P(outLine,parseSerial_Is_OK);
            Serial.println(outLine);
          }else{
            strcpy_P(outLine,parseSerial_Error9);
            Serial.print(outLine);
            Serial.println(inLine);           
          }
		/*按地址解析*/
        }else if('0'<=inLine[5] && inLine[5]<='9'){
          byte ip[4];
          if(parseIPExpr(&inLine[5],ip)){
            strcpy(server_host,&inLine[5]);
            Serial.print("HOST="); Serial.print(server_host); 
            strcpy_P(outLine,parseSerial_Is_OK);
             Serial.println(outLine);
          }else{
            strcpy_P(outLine,parseSerial_Error2);
            Serial.print(outLine);
            Serial.println(inLine);
          }    
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
      
    /* 解析存储路径 */     
    }else if(inLine[0]=='P' && inLine[1]=='A' && inLine[2]=='T' && inLine[3]=='H' && inLine[4]=='='){
      if(config_mode){
        if(0x21<=inLine[5] && inLine[5]<=0x7e){
           if(strlen(&inLine[5])<MAX_SERVER_PATH_LEN){
             strcpy(server_path,&inLine[5]);
             Serial.print("PATH="); Serial.print(server_path); 
             strcpy_P(outLine,parseSerial_Is_OK);
             Serial.println(outLine);
           }else{
             // PROGMEM prog_char parseSerial_Error9[] = "Too Long: ";
             strcpy_P(outLine,parseSerial_Error9);
             Serial.print(outLine);
             Serial.println(inLine);           
           }
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
	  
    /*  解析服务端PORT  */
    }else if(inLine[0]=='P' && inLine[1]=='O' && inLine[2]=='R' && inLine[3]=='T' && inLine[4]=='='){
      if(config_mode){
        unsigned long port_num=atol(&inLine[5]);
        if(0<port_num && port_num<=65535){
           server_port=(unsigned int)port_num;
           Serial.print("PORT="); Serial.print(server_port); 
           strcpy_P(outLine,parseSerial_Is_OK);
           Serial.println(outLine);
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }


	/*解析pointID1*/
	}else if(inLine[0]=='I' && inLine[1]=='D' && inLine[2]=='0' && inLine[3]=='='){
      if(config_mode){
        if( inLine[4]=='h' && inLine[5]=='t' &&
            inLine[6]=='t' && inLine[7]=='p' && 
            inLine[8]==':' && inLine[9]=='/' &&
            inLine[10]=='/' ){
            if(strlen(&inLine[4])<MAX_IEEE1888_POINTID_LEN){
              strcpy(element[0].cid,&inLine[4]);			  
              Serial.print("ID0="); Serial.print(element[0].cid); 
              strcpy_P(outLine,parseSerial_Is_OK);
              Serial.println(outLine);
            }else{
              // PROGMEM prog_char parseSerial_Error9[] = "Too Long: please less than 64";
              strcpy_P(outLine,parseSerial_Error9);
              Serial.print(outLine);
              Serial.println(inLine);           
            }
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }

	/*解析pointID2*/
	}else if(inLine[0]=='I' && inLine[1]=='D' && inLine[2]=='1' && inLine[3]=='='){
      if(config_mode){
        if( inLine[4]=='h' && inLine[5]=='t' &&
            inLine[6]=='t' && inLine[7]=='p' && 
            inLine[8]==':' && inLine[9]=='/' &&
            inLine[10]=='/' ){
            if(strlen(&inLine[4])<MAX_IEEE1888_POINTID_LEN){
              strcpy(element[1].cid,&inLine[4]);
              Serial.print("ID1="); Serial.print(element[1].cid); 
              strcpy_P(outLine,parseSerial_Is_OK);
              Serial.println(outLine);
            }else{
              // PROGMEM prog_char parseSerial_Error9[] = "Too Long: please less than 64";
              strcpy_P(outLine,parseSerial_Error9);
              Serial.print(outLine);
              Serial.println(inLine);           
            }
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
	  
    /* 解析间隔时间 */
    }else if(inLine[0]=='I' && inLine[1]=='N' && inLine[2]=='T' && inLine[3]=='='){
      if(config_mode){
        int len=strlen(&inLine[4]);
        if(0<len && len<5){
          int w=atoi(&inLine[4]);
          if(60<=w && w<=3600){
            MEASURE_INTERVAL=w;
            Serial.print("INT="); Serial.print(w); 
            strcpy_P(outLine,parseSerial_Is_OK);
            Serial.println(outLine);
          }else{
            strcpy_P(outLine,parseSerial_Error2);
            Serial.print(outLine);
            Serial.println(inLine);
          }
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }

    /* 解析Voltage */
    }else if(inLine[0]=='V' && inLine[1]=='='){
      if(config_mode){
        int len=strlen(&inLine[2]);
        if(len==3){
          int w=atoi(&inLine[2]);
          if(100<=w && w<=440){
            voltage_setting=w;
            Serial.print("V="); Serial.print(w); 
            strcpy_P(outLine,parseSerial_Is_OK);
            Serial.println(outLine);
          }else{
            strcpy_P(outLine,parseSerial_Error2);
            Serial.print(outLine);
            Serial.println(inLine);
          }
        }else{
          strcpy_P(outLine,parseSerial_Error2);
          Serial.print(outLine);
          Serial.println(inLine);
        }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }

    /* 解析NTP1（时间服务器地址1）*/
    }else if(inLine[0]=='N' && inLine[1]=='T' && inLine[2]=='P' && inLine[3]=='1' && inLine[4]=='='){
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[5],ip)){
         char tmp[22];
         // PROGMEM prog_char parseSerial_NTP1_Is_Format[] = "NTP1=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_NTP1_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         ntp_server1[0]=ip[0]; ntp_server1[1]=ip[1]; ntp_server1[2]=ip[2]; ntp_server1[3]=ip[3];
         Serial.print(tmp);          
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
      
    /* 解析NTP2（时间服务器地址2）*/
    }else if(inLine[0]=='N' && inLine[1]=='T' && inLine[2]=='P' && inLine[3]=='2' && inLine[4]=='='){
      if(config_mode){
       byte ip[4];
       if(parseIPExpr(&inLine[5],ip)){
         char tmp[22];
         // PROGMEM prog_char parseSerial_NTP2_Is_Format[] = "NTP2=%d.%d.%d.%d";
         strcpy_P(outLine,parseSerial_NTP2_Is_Format);
         sprintf(tmp,outLine,ip[0],ip[1],ip[2],ip[3]);
         ntp_server2[0]=ip[0]; ntp_server2[1]=ip[1]; ntp_server2[2]=ip[2]; ntp_server2[3]=ip[3];
         Serial.print(tmp); 
         strcpy_P(outLine,parseSerial_Is_OK);
         Serial.println(outLine);
       }else{
         strcpy_P(outLine,parseSerial_Error2);
         Serial.print(outLine);
         Serial.println(inLine);
       }
      }else{
        strcpy_P(outLine,parseSerial_Error3);
        Serial.println(outLine);
      }
     
    /* 解析conf指令 */
    }else if(  inLine[0]=='c' 
            && inLine[1]=='o'
            && inLine[2]=='n'
            && inLine[3]=='f'
            && nInLine==4){
              
      if(config_mode){
        // PROGMEM prog_char parseSerial_Error4[] = "Already in configuration mode.";
        strcpy_P(outLine,parseSerial_Error4);
        Serial.println(outLine);
      }else{
        config_mode=1;
		Serial.println("welcome to configure!!!");
      }
      
    /* 解析show指令 */
    }else if(  inLine[0]=='s' 
            && inLine[1]=='h'
            && inLine[2]=='o'
            && inLine[3]=='w'
            && nInLine==4){
      
      char tmp[60];

      strcpy_P(outLine,parseSerial_show_Message00);
      Serial.println(outLine); 
      
      /* 显示MAC */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);      
      strcpy_P(outLine,parseSerial_MAC_Is_Format);
      sprintf(tmp,outLine,local_mac[0],local_mac[1],local_mac[2],local_mac[3],local_mac[4],local_mac[5]);
      Serial.println(tmp);
      
      /* 显示DHCP */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);      
      Serial.print("DHCP=");
      if(dhcp_e){
        Serial.println("true");
      }else{
        Serial.println("false");
      }
      
      /* 显示IP */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);      
      strcpy_P(outLine,parseSerial_IP_Is_Format);
      sprintf(tmp,outLine,local_ip4[0],local_ip4[1],local_ip4[2],local_ip4[3]);
      Serial.println(tmp);
      
      /* 显示NM */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);      
      strcpy_P(outLine,parseSerial_NM_Is_Format);
      sprintf(tmp,outLine,local_netmask[0],local_netmask[1],local_netmask[2],local_netmask[3]);
      Serial.println(tmp);
      
      /* 显示GW */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      strcpy_P(outLine,parseSerial_GW_Is_Format);
      sprintf(tmp,outLine,local_gateway[0],local_gateway[1],local_gateway[2],local_gateway[3]);
      Serial.println(tmp);

      /* 显示DNS */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      strcpy_P(outLine,parseSerial_DNS_Is_Format);
      sprintf(tmp,outLine,dns_addr[0],dns_addr[1],dns_addr[2],dns_addr[3]);
      Serial.println(tmp);
      Serial.println();
      
      /* 显示IEEE1888相关信息 */
      strcpy_P(outLine,parseSerial_show_Message01);
      Serial.println(outLine); 
            
      /* 显示HOST */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      Serial.print("HOST=");
      Serial.println(server_host);
      
      /* 显示存储路径PATH */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      Serial.print("PATH=");
      Serial.println(server_path);
      
      /* 显示服务端端口PORT */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      sprintf(tmp,"PORT=%ld",(unsigned long)server_port);
      Serial.println(tmp);


	/*循环打印出两个pointID*/
	  for(int k=0;k<count;k++){
        strcpy_P(outLine,parseSerial_Indent);
        Serial.print(outLine);
		Serial.print("ID(");
		Serial.print(k);
		Serial.print("):");
		Serial.println(element[k].cid); 		
	  }	  
	  
      /* 显示INT */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      Serial.print("INT=");
      Serial.println(MEASURE_INTERVAL);
	  
      /* 显示V */
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      Serial.print("V=");
      Serial.println(voltage_setting);
      Serial.println();

      /* 显示NTP服务端信息*/
      strcpy_P(outLine,parseSerial_show_Message02);
      Serial.println(outLine); 
      
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      strcpy_P(outLine,parseSerial_NTP1_Is_Format);
      sprintf(tmp,outLine,ntp_server1[0],ntp_server1[1],ntp_server1[2],ntp_server1[3]);
      Serial.println(tmp);
      
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      strcpy_P(outLine,parseSerial_NTP2_Is_Format);
      sprintf(tmp,outLine,ntp_server2[0],ntp_server2[1],ntp_server2[2],ntp_server2[3]);
      Serial.println(tmp);
      Serial.println();

      /* 显示上报状态 */
      strcpy_P(outLine,parseSerial_show_Message05);
      Serial.println(outLine); 
      strcpy_P(outLine,parseSerial_Indent);
      Serial.print(outLine);
      if(ieee1888upload_status==-1){
        // "STAT=NONE"
        strcpy_P(tmp,parseSerial_STAT_NONE_Format);
      }else if(ieee1888upload_status==IEEE1888_UPLOAD_OK){
        // "STAT=OK"
        strcpy_P(tmp,parseSerial_STAT_OK_Format);        
      }else{
        // "STAT=NG(%s)"
        strcpy_P(outLine,parseSerial_STAT_NG_Format);
        if(ieee1888upload_status==IEEE1888_UPLOAD_CONNFAIL){
          sprintf(tmp,outLine,"SOCK");
        }else if(ieee1888upload_status==IEEE1888_UPLOAD_HTTPERR){          
          sprintf(tmp,outLine,"HTTP");
        }else if(ieee1888upload_status==IEEE1888_UPLOAD_IEEE1888ERR){          
          sprintf(tmp,outLine,"1888");
        }else if(ieee1888upload_status==-2){
          sprintf(tmp,outLine,"NTP");
        }else{
          sprintf(tmp,outLine,"?");          
        }
      }
      Serial.println(tmp);
      Serial.println();

    // 解析exit指令
    }else if(  inLine[0]=='e' 
            && inLine[1]=='x'
            && inLine[2]=='i'
            && inLine[3]=='t'
            && nInLine==4){
      
      if(config_mode){
          config_mode=0;
      }else{
        strcpy_P(outLine,parseSerial_Error6);
        Serial.println(outLine);
      }

    //解析save指令
    }else if(  inLine[0]=='s'
      && inLine[1]=='a'
      && inLine[2]=='v'
      && inLine[3]=='e'
      && nInLine==4){

      saveConfig();
      Serial.println("Save ... OK");
      
    // 解析help/?指令
    }else if( (  inLine[0]=='h'
         && inLine[1]=='e'
         && inLine[2]=='l'
         && inLine[3]=='p'
         && nInLine==4     )
         || 
         (  inLine[0]=='?'
         && nInLine==1      ) ){
           
      int k;
      for(k=0;k<30;k++){
        strcpy_P(outLine, (char*)pgm_read_word(&(parseSerial_help_Message_List[k])));
        Serial.println(outLine);
      }
    }else if(nInLine==0){ 
    }else{
      strcpy_P(outLine,parseSerial_Error7);
      Serial.print(outLine);
      Serial.println(inLine);
    }
    
    nInLine=0;
  }
  
  if(is_newline){
    if(config_mode){
       Serial.print("# "); 
    }else{      
       Serial.print("> "); 
    }
  }
}

/*根据串口输入生成MAC地址
   return 0 if failed
   return 1 if success  
*/   
byte parseMACExpr(char* str_mac, byte* mac){
  int i;
  for(i=0;i<6;i++){
    byte front=*(str_mac+i*3);
    byte back=*(str_mac+i*3+1);
    byte separator=*(str_mac+i*3+2);
    if(!((i<5 && (separator=='-' || separator==':')) || (i==5 && separator==0))){
       return 0;
    } 
    
    if('A'<=front && front<='F'){
       front=(front-'A')+10;
    }else if('a'<=front && front<='f'){
       front=(front-'a')+10;
    }else if('0'<=front && front<='9'){
       front=(front-'0');
    }else{
       return 0;
    }
    if('A'<=back && back<='F'){
       back=(back-'A')+10;
    }else if('a'<=back && back<='f'){
       back=(back-'a')+10;
    }else if('0'<=back && back<='9'){
       back=(back-'0');
    }else{
       return 0; 
    }
    *(mac+i)=(front<<4)+back;
  }
  return 1;
}

/*根据串口输入生成IP地址
   return 0 if failed
   return 1 if success
*/   
byte parseIPExpr(char* str_addr, byte* addr){
  int i,s,e;
  s=0; e=0;
  char c;
  char small_buf[4];
  
  for(i=0;i<4;i++){
    for(e=s;e<16;e++){
      c=*(str_addr+e);
      if('0'<=c && c<='9'){
        if(e>=s+4){
          return 0;
        }
        small_buf[e-s]=c;
      }else if((s<e && c=='.' && i<3) || ((c==0x00 || c==0x0d) && i==3)){
        small_buf[e-s]=0;
        break;
      }else{
         return 0;
      }
    }
    s=e+1;
    int intValue=atoi(small_buf);
    if(intValue!=(byte)intValue){
      return 0; 
    }
    *(addr+i)=(byte)intValue;
    if(e>=16 || (i!=3 && (c==0x00 || c==0x0d))){
      return 0;
    }
  }
  return 1;
}

// 将配置数据写入EEPROM
void saveConfig(){

   int i;
   EEPROM.write(0,0xaa);
   EEPROM.write(1,0x55);
   EEPROM.write(2,FIRMWARE_VERSION);

   for(i=0; i<6; i++){
	EEPROM.write(i+3,local_mac[i]);
   }  
   if(dhcp_e){
     EEPROM.write(9,0x0ff);
   }else{
     EEPROM.write(9,0);
   }
   for(i=0;i<4;i++){
      EEPROM.write(i+10,local_ip4[i]);
   }
   for(i=0;i<4;i++){
      EEPROM.write(i+14,local_gateway[i]);
   }
   for(i=0;i<4;i++){
     EEPROM.write(i+18,local_netmask[i]);
   }
   for(i=0;i<4;i++){
      EEPROM.write(i+22,ntp_server1[i]);
   }
   for(i=0;i<4;i++){
      EEPROM.write(i+26,ntp_server2[i]);
   }
   for(i=0;i<4;i++){
      EEPROM.write(i+30,dns_addr[i]);
   }
   EEPROM.write(34,MEASURE_INTERVAL>>8);
   EEPROM.write(35,MEASURE_INTERVAL&0xff);
   EEPROM.write(36,voltage_setting>>8);
   EEPROM.write(37,voltage_setting&0xff);

   for(i=0;i<MAX_SERVER_HOST_LEN;i++){
      EEPROM.write(i+38,server_host[i]);
   }
   for(i=0;i<MAX_SERVER_PATH_LEN;i++){
      EEPROM.write(i+(38+MAX_SERVER_HOST_LEN),server_path[i]);
   }
   EEPROM.write(38+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN,server_port&0xff);
   EEPROM.write(39+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN,server_port/0x100);


   for(i=0; i<MAX_IEEE1888_POINTID_LEN; i++){
	  EEPROM.write(i+(40+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN),element[0].cid[i]);  //写入第一个pointID
   }  
   
   for(i=0; i<MAX_IEEE1888_POINTID_LEN; i++){
	  EEPROM.write(i+(40+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN+MAX_IEEE1888_POINTID_LEN),element[1].cid[i]);  //写入第二个pointID
   }

}

// 从EEPROM读取设置数据
void loadConfig(){
   int i;
   if(EEPROM.read(0)!=0xaa || EEPROM.read(1)!=0x55 || EEPROM.read(2) != FIRMWARE_VERSION) {
      saveConfig();
      for(i=0;i<24;i++){
        EEPROM.write(384+i,0);
      }
      return ;
   } 
  
   for(i=0;i<6;i++){
      local_mac[i]=EEPROM.read(i+3);
   }
   if(EEPROM.read(9)!=0){
     dhcp_e=true;
   }else{
     dhcp_e=false; 
   }
   for(i=0;i<4;i++){
      local_ip4[i]=EEPROM.read(i+10);
   }
   for(i=0;i<4;i++){
      local_gateway[i]=EEPROM.read(i+14);
   }
   for(i=0;i<4;i++){
      local_netmask[i]=EEPROM.read(i+18);
   }
   for(i=0;i<4;i++){
      ntp_server1[i]=EEPROM.read(i+22);
   }
   for(i=0;i<4;i++){
      ntp_server2[i]=EEPROM.read(i+26);
   }
   for(i=0;i<4;i++){
      dns_addr[i]=EEPROM.read(i+30);
   }
   MEASURE_INTERVAL=(EEPROM.read(34)<<8)|EEPROM.read(35);
   voltage_setting=(EEPROM.read(36)<<8)|EEPROM.read(37);
      
   for(i=0;i<MAX_SERVER_HOST_LEN;i++){
      server_host[i]=EEPROM.read(i+38);
   }
   for(i=0;i<MAX_SERVER_PATH_LEN;i++){
      server_path[i]=EEPROM.read(i+(38+MAX_SERVER_HOST_LEN));
   }

   server_port=EEPROM.read(38+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN)
              +(EEPROM.read(39+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN)*0x100);
 
   for(i=0; i<MAX_IEEE1888_POINTID_LEN; i++){
	  element[0].cid[i]= EEPROM.read(i+(40+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN));  //读取第一个pointID
   }
      
   for(i=0; i<MAX_IEEE1888_POINTID_LEN; i++){
	  element[1].cid[i] = EEPROM.read(i+(40+MAX_SERVER_HOST_LEN+MAX_SERVER_PATH_LEN+MAX_IEEE1888_POINTID_LEN));  //读取第二个pointID
   } 
}

const int NTP_MAX_WAITCOUNT = 3;    // wait count to receive response
const int NTP_PACKET_SIZE = 48;

/*-------- NTP code ----------*/
unsigned long getNtpTime()
{
  byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets 
  char n, i;
  
  while (udp.available()) { 
    udp.read(packetBuffer, NTP_PACKET_SIZE);
  }

  for (n = 0; n < NTP_MAX_WAITCOUNT; n++) {
    wdt_reset();
    sendNTPpacket(ntp_server1, packetBuffer);

    delay(500);
    if (!udp.available()) { 

      continue;
    }
    memset(packetBuffer, 0, NTP_PACKET_SIZE); 
    udp.read(packetBuffer, NTP_PACKET_SIZE);

    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);  
    unsigned long secsSince1900 = highWord << 16 | lowWord;
    if (secsSince1900 == 0) {
    continue;
    }
    unsigned long epoch = secsSince1900 - 2208988800UL + timezoneOffset;
    while (udp.available()) { 
      udp.read(packetBuffer, NTP_PACKET_SIZE);
    }  
    return(epoch);
  }
  return(0);
}
unsigned long sendNTPpacket(byte *address, byte *packetBuffer)
{ 
  memset(packetBuffer, 0, NTP_PACKET_SIZE); 
  packetBuffer[0] = 0b11100011;
  packetBuffer[1] = 0;
  packetBuffer[2] = 6;
  packetBuffer[3] = 0xEC;
  packetBuffer[12]  = 49; 
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;
  
  udp.begin(8888);
  udp.beginPacket(address, 123);

  int s = udp.write(packetBuffer, NTP_PACKET_SIZE);
  udp.endPacket();
  
}
// end of code

